extern crate dns_parser;
extern crate etherparse;
extern crate http_muncher;
extern crate iptables;
extern crate libc;
extern crate lettre;
extern crate lettre_email;

use std::net::IpAddr;
use std::net::Ipv4Addr;
use std::net::Ipv6Addr;
use std::sync::mpsc;

use lettre_email::Email;
use lettre::smtp::authentication::Credentials;
use lettre::{SmtpClient, Transport};
use uuid::Uuid;

use hyper::{Body, Request, Response, Client};

use super::models;
pub use models::FirewallClient;
pub use models::Policy;
pub use models::Rule;
use models::MatchStatus;
use super::tshark::TsharkRecord;

use super::lib;

pub fn run(nfqueue_rx: mpsc::Receiver<Vec<u8>>, nfqueue_tx: mpsc::Sender<nfqueue::Verdict>, tshark_rx: mpsc::Receiver<TsharkRecord>,
        get_client_rx: mpsc::Receiver<FirewallClient>, get_client_tx: mpsc::Sender<Option<FirewallClient>>, put_client_rx: mpsc::Receiver<FirewallClient>,
        get_policy_rx: mpsc::Receiver<Policy>, get_policy_tx: mpsc::Sender<Option<Policy>>, put_policy_rx: mpsc::Receiver<Policy>) {
    let mut firewall = Firewall::new();
    info!("Firewall datastore entering main loop.");
    loop {
        match nfqueue_rx.try_recv() {
            Ok(message) => {
                debug!("Datastore received packet from nfqueue.");
                if nfqueue_tx.send(firewall.process_nfqueue_packet(message)).is_err() {
                    info!("Firewall datastore exiting main loop.");
                    return;
                }
            }
            Err(e) => {
                match e {
                    mpsc::TryRecvError::Empty => {}
                    mpsc::TryRecvError::Disconnected => {
                        info!("Firewall datastore exiting main loop.");
                        return;
                    }
                }
            }
        }
        match get_client_rx.try_recv() {
            Ok(client) => {
                debug!("Advertisement get client from datastore.");
                match firewall.get_client(client.ipv4_addr, client.ipv6_addr, client.hostname) {
                    Some(client) => {
                        if get_client_tx.send(Some(client.clone())).is_err() {
                            info!("Firewall datastore exiting main loop.");
                            return;
                        }
                    }
                    None => {
                        if get_client_tx.send(None).is_err() {
                            info!("Firewall datastore exiting main loop.");
                            return;
                        }
                    }
                }
            }
            Err(e) => {
                match e {
                    mpsc::TryRecvError::Empty => {}
                    mpsc::TryRecvError::Disconnected => {
                        info!("Firewall datastore exiting main loop.");
                        return;
                    }
                }
            }
        }
        match put_client_rx.try_recv() {
            Ok(client) => {
                debug!("Advertisement add client to datastore.");
                firewall.register_client(client);
            }
            Err(e) => {
                match e {
                    mpsc::TryRecvError::Empty => {}
                    mpsc::TryRecvError::Disconnected => {
                        info!("Firewall datastore exiting main loop.");
                        return;
                    }
                }
            }
        }
        match get_policy_rx.try_recv() {
            Ok(policy) => {
                debug!("Advertisement get policy from datastore.");
                match firewall.get_policy(policy.uuid, policy.revision, &policy.client) {
                    Some(policy) => {
                        if get_policy_tx.send(Some(policy.clone())).is_err() {
                            info!("Firewall datastore exiting main loop.");
                            return;
                        }
                    }
                    None => {
                        if get_policy_tx.send(None).is_err() {
                            info!("Firewall datastore exiting main loop.");
                            return;
                        }
                    }
                }
            }
            Err(e) => {
                match e {
                    mpsc::TryRecvError::Empty => {}
                    mpsc::TryRecvError::Disconnected => {
                        info!("Firewall datastore exiting main loop.");
                        return;
                    }
                }
            }
        }
        match put_policy_rx.try_recv() {
            Ok(policy) => {
                debug!("Advertisement add policy to datastore.");
                firewall.register_policy(policy);
            }
            Err(e) => {
                match e {
                    mpsc::TryRecvError::Empty => {}
                    mpsc::TryRecvError::Disconnected => {
                        info!("Firewall datastore exiting main loop.");
                        return;
                    }
                }
            }
        }
        match tshark_rx.try_recv() {
            Ok(tshark_record) => {
                debug!("Datastore received hostname from tshark.");
                let mut ipv4_src: Option<Ipv4Addr> = None;
                let mut ipv6_src: Option<Ipv6Addr> = None;
                match tshark_record.src {
                    IpAddr::V4(addr) => {
                        ipv4_src = Some(addr);
                    }
                    IpAddr::V6(addr) => {
                        ipv6_src = Some(addr);
                    }
                }
                let mut ipv4_dst: Option<Ipv4Addr> = None;
                let mut ipv6_dst: Option<Ipv6Addr> = None;
                match tshark_record.dst {
                    IpAddr::V4(addr) => {
                        ipv4_dst = Some(addr);
                    }
                    IpAddr::V6(addr) => {
                        ipv6_dst = Some(addr);
                    }
                }
                let flow = firewall.get_flow(ipv4_src, ipv4_dst, ipv6_src, ipv6_dst, 6, tshark_record.src_port, tshark_record.dst_port);
                match flow {
                    Some(f) => {
                        unsafe {
                            (*f).http_host = Some(tshark_record.hostname);
                        }
                    }
                    None => {
                        warn!("Tshark record for flow we don't know about.");
                    }
                }
            }
            Err(e) => {
                match e {
                    mpsc::TryRecvError::Empty => {}
                    mpsc::TryRecvError::Disconnected => {
                        info!("Firewall datastore exiting main loop.");
                        return;
                    }
                }
            }
        }
    }
}

struct HttpParserHandler {
    drop_packet: bool,
    current_header: String,
}

impl HttpParserHandler {
    pub fn new() -> HttpParserHandler {
        HttpParserHandler {
            drop_packet: false,
            current_header: "".to_string(),
        }
    }
}

impl http_muncher::ParserHandler for HttpParserHandler {
    fn on_header_field(&mut self, _parser: &mut http_muncher::Parser, header: &[u8]) -> bool {
        self.current_header = std::str::from_utf8(header).unwrap().to_string();
        return true;
    }

    fn on_header_value(&mut self, _parser: &mut http_muncher::Parser, value: &[u8]) -> bool {
        if self.current_header == "Host" {
            if std::str::from_utf8(value).unwrap() == "example.net" {
                self.drop_packet = true;
            }
            return false;
        }
        return true;
    }
}

pub fn notify_api(message: String) {
    let client = Client::new();
    let mut request = Request::builder()
        .uri(message)
       .header("message", "System Update")
       .body(Body::from("Message Sent"))
       .expect("request builder");
let response = client.request(request);
}

pub fn notify_email(message: String) {
    let email = Email::builder()
            .to("mqpiotnet@gmail.com")
            .from("mqpiotnet@gmail.com")
            .subject("hello world")
            .html("<h1>Hi there</h1>")
            .text(message)
            .build()
            .unwrap();

        let creds = Credentials::new(
            "mqpiotnet@gmail.com".to_string(),
            "wearenumber1".to_string(),
        );

        // Open connection to gmail
        let mut mailer = SmtpClient::new_simple("smtp.gmail.com")
            .unwrap()
            .credentials(creds)
            .transport();

        // Send the email
        let result = mailer.send(email.into());

        if result.is_ok() {
            info!("Email sent");
        } else {
            info!("Could not send email: {:?}", result);
        }

        assert!(result.is_ok());
}

pub struct Firewall {
    flows: Vec<models::Flow>,
}

impl Firewall {
    pub fn new() -> Firewall {
        Firewall {
            flows: Vec::new(),
        }
    }

    pub fn get_client(&mut self, ipv4_addr: Option<Ipv4Addr>, ipv6_addr: Option<Ipv6Addr>, hostname: String) -> Option<models::FirewallClient> {
		let ipv4_str_opt = match ipv4_addr {
										Some(v) => Some(v.to_string()),
										None => None
		};
		let ipv6_str_opt = match ipv6_addr {
										Some(v) => Some(v.to_string()),
										None => None
		};
		match lib::get_client_from_ip(ipv4_str_opt, ipv6_str_opt) {
			Some(v) => {
				// reuse the parameters for the function because they were used for lookup
				Some(models::FirewallClient::new(ipv4_addr, ipv6_addr, v.hostname).clone())
			},
			None => None
		}
	}

    pub fn register_client(&mut self, new_client: models::FirewallClient) {
    	let ipv4_str_opt = match new_client.ipv4_addr {
										Some(v) => Some(v.to_string()),
										None => None
		};
		let ipv6_str_opt = match new_client.ipv6_addr {
										Some(v) => Some(v.to_string()),
										None => None
		};
    	lib::add_client(ipv4_str_opt, ipv6_str_opt, new_client.hostname);
    }

    pub fn policy_exists(&mut self, policy: &models::Policy) -> bool {
        return self.get_policy(policy.uuid.clone(), policy.revision.clone(), &policy.client).is_some();
    }

	/* Get all the policies for this client */
	pub fn get_policies_for_client(&mut self, client: &models::FirewallClient) -> Vec<models::Policy> {
		let mut policies: Vec<models::Policy> = Vec::new();
		// get database client from ipv4 and ipv6 address
		// get client id from database client
		let ipv4_str_opt = match client.ipv4_addr {
										Some(v) => Some(v.to_string()),
										None => None
		};
		let ipv6_str_opt = match client.ipv6_addr {
										Some(v) => Some(v.to_string()),
										None => None
		};
		let database_client = lib::get_client_from_ip(ipv4_str_opt, ipv6_str_opt);
		if !database_client.is_some() {
			// if the client was not found, return the empty vector
			// this would only occur if we are seeing traffic from a client that has no policies, therefore is not in the clients table.
			return policies;
		}
		let client_id = database_client.unwrap().client_id;
		// get all policyclients connected to client_id from policyclients table
		let policyclients = lib::get_policy_clients_from_client_id(client_id);
		
		// get all uuid from the policyclients
		for policyclient in policyclients {
			let uuid = policyclient.policy_uuid;
			// get revision number from policy table
			let rev_num = lib::get_policy(uuid.clone()).unwrap().rev_num;
			
			// use get_policy() function below to get a policy from uuid, rev, client.
			let policy = self.get_policy(Uuid::parse_str(&uuid).unwrap(), rev_num as u32, client);
			// unwrap option policy from below, add it to the vector
			policies.push(policy.unwrap());
			// repeat for all other policies
		}
		// return vector
		return policies;
	}
	
	/* Get the policy from the database */
    pub fn get_policy(&mut self, uuid: Uuid, revision: u32, client: &models::FirewallClient) -> Option<models::Policy> {
    	// Set the policy with the data in the database
        let mut new_policy = models::Policy::new(uuid, revision, client.clone());
        
        // set the certificate if there is a policy
        let policy = lib::get_policy(uuid.to_hyphenated().to_string());
        if !policy.is_some() {
        	return None;
        }
        if policy.clone().unwrap().certificate.is_some() {
        	new_policy.set_certificate(policy.unwrap().certificate.unwrap());
        }
        
        // set the tags
        let tags = lib::get_tags_from_policy_uuid(uuid.to_hyphenated().to_string());
        for tag in tags {
        	new_policy.add_tag(tag.tag_name);
        }
        
        // set the rules and notifications
        let rules = lib::get_rules_from_policy_uuid(uuid.to_hyphenated().to_string());
        for rule in rules {
        	let mut new_rule = Rule::new(rule.action);
        	
        	// add selectors to this rule
        	if rule.ipv4_remote_addr.is_some() {
        		new_rule.add_selector("ipv4-remote-addr".to_string(), rule.ipv4_remote_addr.unwrap().to_string());
        	}
        	if rule.ipv6_remote_addr.is_some() {
        		new_rule.add_selector("ipv6-remote-addr".to_string(), rule.ipv6_remote_addr.unwrap().to_string());
        	}
        	if rule.tcp_remote_port.is_some() {
        		new_rule.add_selector("tcp-remote-port".to_string(), rule.tcp_remote_port.unwrap().to_string());
        	}
        	if rule.udp_remote_port.is_some() {
        		new_rule.add_selector("udp-remote-port".to_string(), rule.udp_remote_port.unwrap().to_string());
        	}
        	
        	// add notifications to rule
        	// get notificiation ids for this rule num
        	let rule_notifications = lib::get_rule_notifications(uuid.to_hyphenated().to_string(), rule.rule_num);
			for rule_notification in rule_notifications {
				// get notification from notification id
				let notification_id = rule_notification.notification_id;
				let notification = lib::get_notification(uuid.to_hyphenated().to_string(), notification_id).clone().unwrap(); // there should always be a notification in the notification table if the id was in the rulenotification table
				
				if notification.email.is_some() {
        			new_rule.add_notification("email".to_string(), notification.email.unwrap().to_string());
		    	}
		    	if notification.rest_api.is_some() {
		    		new_rule.add_notification("rest-api".to_string(), notification.rest_api.unwrap().to_string());
		    	}
			}

        	new_policy.add_rule(new_rule);
        }
        
        return Some(new_policy);
    }

	/* Add the policy to the database */
    pub fn register_policy(&mut self, new_policy: models::Policy) {
        // add to policy table
        let p = &new_policy;
        let policy_uuid = (&new_policy.uuid).to_hyphenated().to_string();
        let rev_num = *(&new_policy.revision) as i32;
        let certificate = p.certificate.clone();
        // if this policy is a new revision, delete all old information from the old policy. if it is not, return
        let old_policy = lib::get_policy(policy_uuid.to_string());
        if old_policy.is_some() && rev_num <= old_policy.unwrap().rev_num {
        	return;
        } else {
        	lib::delete_rules(policy_uuid.to_string());
        	lib::delete_notifications(policy_uuid.to_string());
        	lib::delete_rule_notifications(policy_uuid.to_string());
        	lib::delete_tags_from_policy_uuid(policy_uuid.to_string());
        	// TODO should more tables be deleted from?
        }
        lib::update_policy(policy_uuid.to_string(), rev_num, certificate);
        
        // add to tags table
        let tags = &new_policy.tags;
        for tag in tags {
        	lib::add_tag(policy_uuid.to_string(), tag.to_string());
        }
        
        // add to clients table
        let client = &new_policy.client;
        let ipv4_addr = match client.ipv4_addr {
        		Some(v) => Some(v.to_string()),
        		None => None,
        };
        let ipv6_addr = match client.ipv6_addr {
        		Some(v) => Some(v.to_string()),
        		None => None,
        };
        let hostname = &client.hostname;
        let client_id = lib::add_client(ipv4_addr, ipv6_addr, hostname.to_string()).client_id;
        
        // add to policyclients table
        lib::add_policy_client(policy_uuid.to_string(), client_id);
        
        // add to rules, notifications, and rulenotifcations tables
        let rules = &new_policy.rules;
        let mut rule_num = 1;
        let mut notification_id = 1; // notification_id is basically a notification_num
        for rule in rules {
        	let action = &rule.action;
        	let mut ipv4_remote_addr = None;
        	let mut ipv6_remote_addr = None;
        	let mut tcp_remote_port = None;
        	let mut udp_remote_port = None;
        	
        	let selectors = &rule.selectors;
        	for selector in selectors {
        		match selector.my_type() {
        			"ipv4-remote-address" => {ipv4_remote_addr = Some(selector.get_ipv4_address().to_string())},
        			"ipv6-remote-address" => {ipv6_remote_addr = Some(selector.get_ipv6_address().to_string())},
        			"tcp-remote-port" => {tcp_remote_port = Some(selector.get_port() as i32)},
        			"udp-remote-port" => {udp_remote_port = Some(selector.get_port() as i32)},
        			&_ => {},
        		}
        	}
        	
        	lib::update_rule(policy_uuid.to_string(), rule_num, action.to_string(), ipv4_remote_addr,
    				ipv6_remote_addr, tcp_remote_port, udp_remote_port);
    				
        	
        	let notifications = &rule.notifications;
        	for notification in notifications {
        		let mut email = None;
        		let mut rest_api = None;
        		match notification.my_type() {
        			"email" => {email = Some(notification.get_email().to_string())},
        			"rest-api" => {rest_api = Some(notification.get_rest_api().to_string())},
        			&_ => {},
        		}
        		// TODO we are adding extra entries into the notification table, unavoidable?
				// only add the notification if it was a part of this rule
				if email.is_some() || rest_api.is_some() {
					lib::update_notification(policy_uuid.to_string(), notification_id, email, rest_api);
					lib::add_rule_notification(policy_uuid.to_string(), rule_num, notification_id);
					
					notification_id += 1;
				}
        	}
    		
    		rule_num += 1;
        }
    }

    fn has_flow(&self, ipv4_source: Option<Ipv4Addr>, ipv4_dest: Option<Ipv4Addr>,
            ipv6_source: Option<Ipv6Addr>, ipv6_dest: Option<Ipv6Addr>,
            ip_proto: u8, source_port: u16, dest_port: u16) -> bool {
        !self.flows.iter().filter(|f| f.ipv4_source == ipv4_source && f.ipv4_dest == ipv4_dest
            && f.ipv6_source == ipv6_source && f.ipv6_dest == ipv6_dest
            && f.ip_proto == ip_proto && f.source_port == source_port
            && f.dest_port == dest_port).next().is_none()
    }

    fn get_flow(&mut self, ipv4_source: Option<Ipv4Addr>, ipv4_dest: Option<Ipv4Addr>,
            ipv6_source: Option<Ipv6Addr>, ipv6_dest: Option<Ipv6Addr>,
            ip_proto: u8, source_port: u16, dest_port: u16) -> Option<*mut models::Flow> {
        match self.flows.iter_mut().filter(|f| f.ipv4_source == ipv4_source && f.ipv4_dest == ipv4_dest
            && f.ipv6_source == ipv6_source && f.ipv6_dest == ipv6_dest
            && f.ip_proto == ip_proto && f.source_port == source_port
            && f.dest_port == dest_port).next() {
                Some(flow) => Some(flow as *mut models::Flow),
                None => None
            }
    }

    fn new_flow(&self, ipv4_source: Option<Ipv4Addr>, ipv4_dest: Option<Ipv4Addr>,
            ipv6_source: Option<Ipv6Addr>, ipv6_dest: Option<Ipv6Addr>,
            ip_proto: u8, source_port: u16, dest_port: u16) -> models::Flow {
        assert!(!self.has_flow(ipv4_source, ipv4_dest, ipv6_source, ipv6_dest, ip_proto, source_port, dest_port));
        let mut new_flow = models::Flow::new(ipv4_source, ipv4_dest, ipv6_source, ipv6_dest, ip_proto, source_port, dest_port);
        let (client, direction) = self.identify_client(&new_flow);
        new_flow.client = client;
        new_flow.direction = direction;
        unsafe {
            if new_flow.client.is_some() && new_flow.direction.is_some() {
                debug!("New flow {} {}.", (*new_flow.client.as_ref().unwrap()).hostname, new_flow.direction.as_ref().unwrap());
            } else {
                debug!("New flow with no client identified.")
            }
        }
        return new_flow;
    }

    fn register_flow(&mut self, new_flow: models::Flow) -> &mut models::Flow {
        self.flows.push(new_flow);
        return self.flows.last_mut().unwrap();
    }

    fn identify_client(&self, flow: &models::Flow) -> (Option<models::FirewallClient>, Option<models::FlowDirection>) {
        // TODO this is a temporary hack until we determine a spoofing-resistant method to identify clients
        // convert data types into types that can be searched by the database.
        let ipv4_str_opt_in = match flow.ipv4_source {
										Some(v) => Some(v.to_string()),
										None => None
		};
		let ipv6_str_opt_in = match flow.ipv6_source {
										Some(v) => Some(v.to_string()),
										None => None
		};
		let ipv4_str_opt_out = match flow.ipv4_dest {
										Some(v) => Some(v.to_string()),
										None => None
		};
		let ipv6_str_opt_out = match flow.ipv6_dest {
										Some(v) => Some(v.to_string()),
										None => None
		};
		let client_in = lib::get_client_from_ip(ipv4_str_opt_in.clone(), ipv6_str_opt_in.clone());
		let client_out = lib::get_client_from_ip(ipv4_str_opt_out.clone(), ipv6_str_opt_out.clone());
		trace!("FirewallClients in and out: {:?}, {:?}", client_in, client_out);
		
        // TODO inbound and outbound traffic need to be fixed
        if client_in.is_some() { // if inbound, use the source (remote address)
        	// convert database client into client
        	let client = Some(models::FirewallClient::new(flow.ipv4_source, flow.ipv6_source, client_in.unwrap().hostname));
        	unsafe {
		        if (client.clone().unwrap()).ipv4_addr.is_some() && flow.ipv4_source.is_some() && (client.clone().unwrap()).ipv4_addr.unwrap() == flow.ipv4_source.unwrap() {
		            return (client, Some(models::FlowDirection::Outbound));
		        }
		        if (client.clone().unwrap()).ipv6_addr.is_some() && flow.ipv6_source.is_some() && (client.clone().unwrap()).ipv6_addr.unwrap() == flow.ipv6_source.unwrap() {
		            return (client, Some(models::FlowDirection::Outbound));
		        }
		    }
        } else if client_out.is_some() { // if outbound traffic, use the destination (remote address)
        	// convert database client into client
        	let client = Some(models::FirewallClient::new(flow.ipv4_dest, flow.ipv6_dest, client_in.unwrap().hostname));
        	unsafe {
		        if (client.clone().unwrap()).ipv4_addr.is_some() && flow.ipv4_dest.is_some() && (client.clone().unwrap()).ipv4_addr.unwrap() == flow.ipv4_dest.unwrap() {
		            return (client, Some(models::FlowDirection::Inbound));
		        }
		        if (client.clone().unwrap()).ipv6_addr.is_some() && flow.ipv6_dest.is_some() && (client.clone().unwrap()).ipv6_addr.unwrap() == flow.ipv6_dest.unwrap() {
		            return (client, Some(models::FlowDirection::Inbound));
		        }
		    }
        } else {
        	// if there was no match to this flow, return an empty tuple
        	return (None, None);
        }
        return (None, None);
    }

    fn process_policies(&mut self, flow: *mut models::Flow) {
        // Process all policies for the client
        let mut overall_status: MatchStatus = MatchStatus::NoMatch;
        unsafe {
        	let policies = self.get_policies_for_client(&(*flow).client.clone().unwrap());
			trace!("FirewallClient processing policy for {:?}", &(*flow).client.clone().unwrap().ipv4_addr);
            for policy in policies {
                /*if policy.client != (*flow).client.clone().unwrap() {
                    continue; //TODO can remove this now
                }*/
                let bookmark: &usize = (*flow).match_state.get(&policy).unwrap_or(&0);
                let (status, bookmark) = policy.process(flow, bookmark);
                debug!("policy {} {}", policy.uuid, status);
                match status {
                    MatchStatus::NoMatch => {}
                    MatchStatus::NotYetKnown => {
                        match overall_status {
                            MatchStatus::NoMatch => {
                                overall_status = MatchStatus::NotYetKnown;
                            }
                            MatchStatus::NotYetKnown | MatchStatus::MatchPass | MatchStatus::MatchDrop | MatchStatus::MatchReject => {}
                        }
                        (*flow).match_state.insert(policy, bookmark);
                    }
                    MatchStatus::MatchPass => {
                        match overall_status {
                            MatchStatus::NoMatch | MatchStatus::NotYetKnown => {
                                overall_status = MatchStatus::MatchPass;
                            }
                            MatchStatus::MatchPass | MatchStatus::MatchDrop | MatchStatus::MatchReject => {}
                        }
                    }
                    MatchStatus::MatchReject => {
                        match overall_status {
                            MatchStatus::NoMatch | MatchStatus::NotYetKnown | MatchStatus::MatchPass => {
                                overall_status = MatchStatus::MatchReject;
                            }
                            MatchStatus::MatchDrop | MatchStatus::MatchReject => {}
                        }
                    }
                    MatchStatus::MatchDrop => {
                        overall_status = MatchStatus::MatchDrop;
                    }
                }
            }
            (*flow).match_status = overall_status;
        }
    }

    fn process_nfqueue_packet(&mut self, packet: Vec<u8>) -> nfqueue::Verdict {
        let pkt: etherparse::SlicedPacket = match etherparse::SlicedPacket::from_ip(packet.as_slice()) {
            Ok(p) => p,
            Err(_) => {
                // failure to parse the packet
                return nfqueue::Verdict::Drop;
            }
        };
    
        // Parse fields from packet
        let mut ipv4_source: Option<Ipv4Addr> = None;
        let mut ipv4_dest: Option<Ipv4Addr> = None;
        let mut ipv6_source: Option<Ipv6Addr> = None;
        let mut ipv6_dest: Option<Ipv6Addr> = None;
        let mut ip_proto: u8 = 0;
        let mut source_port: u16 = 0;
        let mut dest_port: u16 = 0;
        
        // Parse IPv4 and IPv6 addresses
        let ip_header = &pkt.ip;
        if ip_header.is_some() {
            match ip_header.as_ref().unwrap() {
                etherparse::InternetSlice::Ipv4(ipv4) => {
                    ipv4_source = Some(ipv4.source_addr());
                    ipv4_dest = Some(ipv4.destination_addr());
                    ip_proto = ipv4.protocol();
                }
                etherparse::InternetSlice::Ipv6(ipv6, _ipv6_extension_header) => {
                    ipv6_source = Some(ipv6.source_addr());
                    ipv6_dest = Some(ipv6.destination_addr());
                    ip_proto = ipv6.next_header();
                }
            }
        }
        // Parse TCP and UDP ports
        let transport = &pkt.transport;
        if transport.is_some() {
            match transport.as_ref().unwrap() {
                etherparse::TransportSlice::Tcp(tcp) => {
                    source_port = tcp.source_port();
                    dest_port = tcp.destination_port();
                }
                etherparse::TransportSlice::Udp(udp) => {
                    source_port = udp.source_port();
                    dest_port = udp.destination_port();
                }
            }
        }
    
        let flow: *mut models::Flow;
        match self.get_flow(ipv4_source, ipv4_dest, ipv6_source, ipv6_dest, ip_proto, source_port, dest_port) {
            Some(fl) => {
                flow = fl;
            }
            None => {
                flow = self.register_flow(self.new_flow(ipv4_source, ipv4_dest, ipv6_source, ipv6_dest, ip_proto, source_port, dest_port));
            }
        }
        
        unsafe {
			debug!("Flow: {:?}, {:?}, {:?}, {:?}, {:?}", (*flow).ipv4_source, (*flow).ipv4_dest, (*flow).source_port, (*flow).dest_port, (*flow).ip_proto);
            let client: Option<models::FirewallClient> = (*flow).client.clone();
            match client {
                None => {
                    // the packet is not to/from any client that we know about
                    return nfqueue::Verdict::Accept;
                }
                Some(c) => {
                    match (*flow).match_status {
                        MatchStatus::NotYetKnown => {
                            debug!("packet from client {}", c.hostname);
                            self.process_policies(flow);
                        }
                        _ => {
                            debug!("packet from client {} cached verdict {}", c.hostname, (*flow).match_status);
                        }
                    }
                    // Take action based on the overall result
                    match (*flow).match_status {
                        MatchStatus::NoMatch => {
                            // no rule in any policy matches this packet, default drop
                            return nfqueue::Verdict::Drop;
                        }
                        MatchStatus::NotYetKnown => {
                            // Let packets pass until we know the flow is bad
                            return nfqueue::Verdict::Accept;
                        }
                        MatchStatus::MatchPass => {
                            return nfqueue::Verdict::Accept;
                        }
                        MatchStatus::MatchReject => {
                            // TODO
                            unimplemented!();
                        }
                        MatchStatus::MatchDrop => {
                            return nfqueue::Verdict::Drop;
                        }
                    }
                }
            }
        }
    }
}
