use diesel::prelude::*;
use diesel::pg::PgConnection;
use dotenv::dotenv;
use std::env;

use crate::firewall::models::*;
use crate::firewall::schema::*;


const QUERY_LIMIT: i64 = 100;

pub fn establish_connection() -> PgConnection {
    dotenv().ok();

    let database_url = env::var("DATABASE_URL")
        .expect("DATABASE_URL must be set");
    PgConnection::establish(&database_url)
        .expect(&format!("Error connecting to {}", database_url))
}


/****************************************************************************************
 * Policy
 ****************************************************************************************/

pub fn add_policy(policy_uuid: String, rev_num: i32, certificate: Option<String>) -> DatabasePolicy {
	// Create new diesel connection
	let connection = establish_connection();

	let new_policy = DatabasePolicy {
        policy_uuid: policy_uuid,
    	rev_num: rev_num,
    	certificate: certificate,
    };

	diesel::insert_into(policies::table)
        .values(&new_policy)
        .get_result(&connection)
        .expect("Error saving new database policy")
}



pub fn update_policy(policy_uuid: String, rev_num: i32, certificate: Option<String>) -> DatabasePolicy {
	// Create new diesel connection
	let connection = establish_connection();
	let primary_key = &policy_uuid;

	let new_policy = DatabasePolicy {
        policy_uuid: primary_key.to_string(),
    	rev_num: rev_num,
    	certificate: certificate,
    };

	// Update the policy or add it if it is new. If either error, panic.
	let policy = match diesel::update(policies::table.find(&policy_uuid))
        .set(&new_policy)
        .get_result::<DatabasePolicy>(&connection) {
			Ok(v) => v,
			Err(e) => {
				if e == diesel::result::Error::NotFound {
					// Policy Uuid was not found, add an entry
					debug!("Adding policy: {:?}", &policy_uuid);
					match diesel::insert_into(policies::table)
							.values(&new_policy)
							.get_result(&connection) {
							Ok(v) => v,
							Err(e) => {
								panic!("Unable to add policy: {:?}", e)
							},						
						}
				} else {
					panic!("Unable to update policy: {:?}", e)
				}
			},
		};

    debug!("Updated policy {}", policy.policy_uuid);
	policy
}



pub fn get_policy(policy_uuid: String) -> Option<DatabasePolicy> {
	// Create new diesel connection
	let connection: PgConnection = establish_connection();

	let results = policies::table.filter(policies::policy_uuid.eq(policy_uuid))
        .limit(1)
        .load::<DatabasePolicy>(&connection)
        .expect("Error loading policies");
	
    debug!("Getting {} policies", results.len());
	if results.len() >= 1 {
		let policy = &results[0];
		Some(DatabasePolicy::clone(policy))
	} else {
		None
	}
}



pub fn delete_policy(policy_uuid: String) {
	// Create new diesel connection
	let connection = establish_connection();

	let num_deleted = diesel::delete(policies::table.filter(policies::policy_uuid.like(policy_uuid)))
        .execute(&connection)
        .expect("Error deleting policies");

    debug!("Deleted {} policies", num_deleted);
}


/****************************************************************************************
 * Client
 ****************************************************************************************/

pub fn update_client(client_id: i32, ipv4_addr: Option<String>, ipv6_addr: Option<String>, hostname: String) -> DatabaseClient {
	// Create new diesel connection
	let connection = establish_connection();
	let primary_key = &client_id;

	let new_client = DatabaseClient {
        client_id: *primary_key,
    	ipv4_addr: ipv4_addr,
		ipv6_addr: ipv6_addr,
    	hostname: hostname,
    };

	// Update the client or add it if it is new. If either error, panic.
	let client = match diesel::update(clients::table.find(&client_id))
        .set(&new_client)
        .get_result::<DatabaseClient>(&connection) {
			Ok(v) => v,
			Err(e) => {
				if e == diesel::result::Error::NotFound {
					// Client id was not found, add an entry
					trace!("Adding client: {:?}", &client_id);
					match diesel::insert_into(clients::table)
							.values(&new_client)
							.get_result(&connection) {
							Ok(v) => v,
							Err(e) => {
								panic!("Unable to add client: {:?}", e)
							},						
						}
				} else {
					panic!("Unable to update client: {:?}", e)
				}
			},
		};

    trace!("Updated client {}", client.client_id);
	client
}



pub fn add_client(ipv4_addr: Option<String>, ipv6_addr: Option<String>, hostname: String) -> DatabaseClient {
	// Create new diesel connection
	let connection = establish_connection();

	let new_client = NewDatabaseClient {
    	ipv4_addr: ipv4_addr.clone(),
		ipv6_addr: ipv6_addr.clone(),
    	hostname: hostname.clone(),
    };

	// add a new client if nothing is already there
	let results = clients::table.filter(clients::ipv4_addr.eq(ipv4_addr.clone()).or(clients::ipv4_addr.is_null().and(!ipv4_addr.is_some())))
								.filter(clients::ipv6_addr.eq(ipv6_addr.clone()).or(clients::ipv6_addr.is_null().and(!ipv6_addr.is_some())))
								.filter(clients::hostname.eq(hostname))
        .limit(QUERY_LIMIT)
        .load::<DatabaseClient>(&connection)
        .expect("Error loading clients");

	if results.len() > 0 {
		trace!("Client already added");
		results[0].clone()
	} else {
		trace!("Adding 1 client");
		diesel::insert_into(clients::table)
		    .values(&new_client)
		    .get_result(&connection)
		    .expect("Error saving new database client")
	}
}



pub fn get_client(client_id: i32) -> Option<DatabaseClient> {
	// Create new diesel connection
	let connection = establish_connection();

	let results = clients::table.filter(clients::client_id.eq(client_id))
        .limit(1)
        .load::<DatabaseClient>(&connection)
        .expect("Error loading clients");

    trace!("Getting {} clients from id", results.len());
	if results.len() >= 1 {
		let client = &results[0];
		Some(DatabaseClient::clone(client))
	} else {
		None
	}
}



pub fn get_client_from_ip(ipv4_addr: Option<String>, ipv6_addr: Option<String>) -> Option<DatabaseClient> {
	// Create new diesel connection
	let connection = establish_connection();

	let results = clients::table.filter(clients::ipv4_addr.eq(ipv4_addr.clone()).or(clients::ipv4_addr.is_null().and(!ipv4_addr.is_some())))
								.filter(clients::ipv6_addr.eq(ipv6_addr.clone()).or(clients::ipv6_addr.is_null().and(!ipv6_addr.is_some())))
        .limit(1)
        .load::<DatabaseClient>(&connection)
        .expect("Error loading clients");

    trace!("Getting {} clients from ip address {:?}, {:?}", results.len(), ipv4_addr, ipv6_addr);
	if results.len() >= 1 {
		let client = &results[0];
		Some(DatabaseClient::clone(client))
	} else {
		None
	}
}



pub fn delete_client(client_id: i32) {
	// Create new diesel connection
	let connection = establish_connection();

	let num_deleted = diesel::delete(clients::table.filter(clients::client_id.eq(client_id)))
        .execute(&connection)
        .expect("Error deleting clients");

    trace!("Deleted {} clients", num_deleted);
}


/****************************************************************************************
 * PolicyClient
 ****************************************************************************************/

pub fn add_policy_client(policy_uuid: String, client_id: i32) -> DatabasePolicyClient {
	// Create new diesel connection
	let connection = establish_connection();
	let primary_key_policy_uuid = &policy_uuid;

	let new_policy_client = DatabasePolicyClient {
        policy_uuid: primary_key_policy_uuid.to_string(),
    	client_id: client_id,
    };

	// add a new policy if nothing is already there
	let results = policyclients::table.filter(policyclients::policy_uuid.eq(policy_uuid))
									  .filter(policyclients::client_id.eq(client_id))
        .limit(QUERY_LIMIT)
        .load::<DatabasePolicyClient>(&connection)
        .expect("Error loading policyclients");

	if results.len() > 0 {
		trace!("Policy client already added");
		new_policy_client	
	} else {
		trace!("Adding 1 policyclient");
		diesel::insert_into(policyclients::table)
		    .values(&new_policy_client)
		    .get_result(&connection)
		    .expect("Error saving new database policyclient")
	}
}



pub fn get_policy_clients_from_policy_uuid(policy_uuid: String) -> Vec<DatabasePolicyClient> {
	// Create new diesel connection
	let connection = establish_connection();

	let results = policyclients::table.filter(policyclients::policy_uuid.eq(policy_uuid))
        .limit(QUERY_LIMIT)
        .load::<DatabasePolicyClient>(&connection)
        .expect("Error loading policyclients");

    trace!("Getting {} policyclients", results.len());
	results
}



pub fn get_policy_clients_from_client_id(client_id: i32) -> Vec<DatabasePolicyClient> {
	// Create new diesel connection
	let connection = establish_connection();

	let results = policyclients::table.filter(policyclients::client_id.eq(client_id))
        .limit(QUERY_LIMIT)
        .load::<DatabasePolicyClient>(&connection)
        .expect("Error loading policyclients");

    trace!("Getting {} policyclients", results.len());
	results
}



pub fn delete_policy_client(policy_uuid: String, client_id: i32) {
	// Create new diesel connection
	let connection = establish_connection();

	let num_deleted = diesel::delete(policyclients::table.filter(policyclients::client_id.eq(client_id))
														 .filter(policyclients::policy_uuid.eq(policy_uuid)))
        .execute(&connection)
        .expect("Error deleting policyclients");

    trace!("Deleted {} policyclients", num_deleted);
}


/****************************************************************************************
 * Rule
 ****************************************************************************************/

pub fn update_rule(policy_uuid: String, rule_num: i32, action: String, ipv4_remote_addr: Option<String>,
    				ipv6_remote_addr: Option<String>, tcp_remote_port: Option<i32>, udp_remote_port: Option<i32>) -> DatabaseRule {
	// Create new diesel connection
	let connection = establish_connection();
	let primary_key_policy_uuid = &policy_uuid;
	let primary_key_rule_num = &rule_num;

	let new_rule = DatabaseRule {
		policy_uuid: primary_key_policy_uuid.to_string(),
    	rule_num: *primary_key_rule_num,
    	action: action,
    	ipv4_remote_addr: ipv4_remote_addr,
    	ipv6_remote_addr: ipv6_remote_addr,
    	tcp_remote_port: tcp_remote_port,
    	udp_remote_port: udp_remote_port,
    };

	// Update the rule or add it if it is new. If either error, panic.
	let rule = match diesel::update(rules::table.find(new_rule.id()))
        .set(&new_rule)
        .get_result::<DatabaseRule>(&connection) {
			Ok(v) => v,
			Err(e) => {
				if e == diesel::result::Error::NotFound {
					// Client id was not found, add an entry
					trace!("Adding rule: {:?}, {:?}", &policy_uuid, rule_num);
					match diesel::insert_into(rules::table)
							.values(&new_rule)
							.get_result(&connection) {
							Ok(v) => v,
							Err(e) => {
								panic!("Unable to add rule: {:?}", e)
							},						
						}
				} else {
					panic!("Unable to update rule: {:?}", e)
				}
			},
		};

    trace!("Updated rule {}, {}", rule.policy_uuid, rule.rule_num);
	rule
}



pub fn get_rule(policy_uuid: String, rule_num: i32) -> Option<DatabaseRule> {
	// Create new diesel connection
	let connection = establish_connection();
	
	let mut r: DatabaseRule = DatabaseRule {
										policy_uuid: policy_uuid,
										rule_num: rule_num,
										action: "".to_string(),
										ipv4_remote_addr: None,
										ipv6_remote_addr: None,
										tcp_remote_port: None,
										udp_remote_port: None
									};	
	
    trace!("Getting {:?} rule", r);

	match rules::table.find(r.id()).first::<DatabaseRule>(&connection) {
		Ok(v) => Some(v),
		Err(v) => None,
	}
}



pub fn get_rules_from_policy_uuid(policy_uuid: String) -> Vec<DatabaseRule> {
	// Create new diesel connection
	let connection = establish_connection();

	let results = rules::table.filter(rules::policy_uuid.eq(policy_uuid))
        .limit(QUERY_LIMIT)
        .load::<DatabaseRule>(&connection)
        .expect("Error loading rules");

    trace!("Getting {} rules", results.len());
	results
}



pub fn delete_rule(policy_uuid: String, rule_num: i32) {
	// Create new diesel connection
	let connection = establish_connection();

	let num_deleted = diesel::delete(rules::table.filter(rules::policy_uuid.eq(policy_uuid))
												 .filter(rules::rule_num.eq(rule_num)))
        .execute(&connection)
        .expect("Error deleting rules");

    trace!("Deleted {} rules", num_deleted);
}

pub fn delete_rules(policy_uuid: String) {
	// Create new diesel connection
	let connection = establish_connection();

	let num_deleted = diesel::delete(rules::table.filter(rules::policy_uuid.eq(policy_uuid)))
        .execute(&connection)
        .expect("Error deleting rules");

    trace!("Deleted {} rules", num_deleted);
}


/****************************************************************************************
 * Notification
 ****************************************************************************************/

pub fn update_notification(policy_uuid: String, notification_id: i32, email: Option<String>, rest_api: Option<String>) -> DatabaseNotification {
	// Create new diesel connection
	let connection = establish_connection();
	let primary_key_policy_uuid = &policy_uuid;
	let primary_key_notification_id = &notification_id;

	let new_notification = DatabaseNotification {
		policy_uuid: primary_key_policy_uuid.to_string(),
    	notification_id: *primary_key_notification_id,
		email: email,
		rest_api: rest_api,
    };

	// Update the notification or add it if it is new. If either error, panic.
	let notification = match diesel::update(notifications::table.find(new_notification.id()))
        .set(&new_notification)
        .get_result::<DatabaseNotification>(&connection) {
			Ok(v) => v,
			Err(e) => {
				if e == diesel::result::Error::NotFound {
					// Client id was not found, add an entry
					trace!("Adding notification: {:?}, {:?}", &policy_uuid, notification_id);
					match diesel::insert_into(notifications::table)
							.values(&new_notification)
							.get_result(&connection) {
							Ok(v) => v,
							Err(e) => {
								panic!("Unable to add notification: {:?}", e)
							},						
						}
				} else {
					panic!("Unable to update notification: {:?}", e)
				}
			},
		};

    trace!("Updated notification {}, {}", notification.policy_uuid, notification.notification_id);
	notification
}



pub fn get_notification(policy_uuid: String, notification_id: i32) -> Option<DatabaseNotification> {
	// Create new diesel connection
	let connection = establish_connection();
	
	let mut n: DatabaseNotification = DatabaseNotification {
													policy_uuid: policy_uuid,
													notification_id: notification_id,
													email: Some("".to_string()),
													rest_api: Some("".to_string()),
												};	
	
    trace!("Getting {:?} notification", n);

	match notifications::table.find(n.id()).first::<DatabaseNotification>(&connection) {
		Ok(v) => Some(v),
		Err(v) => None,
	}
}



pub fn get_notifications_from_policy_uuid(policy_uuid: String) -> Vec<DatabaseNotification> {
	// Create new diesel connection
	let connection = establish_connection();

	let results = notifications::table.filter(notifications::policy_uuid.eq(policy_uuid))
        .limit(QUERY_LIMIT)
        .load::<DatabaseNotification>(&connection)
        .expect("Error loading notifications");

    trace!("Getting {} notifications", results.len());
	results
}



pub fn delete_notifications(policy_uuid: String) {
	// Create new diesel connection
	let connection = establish_connection();

	let num_deleted = diesel::delete(notifications::table.filter(notifications::policy_uuid.eq(policy_uuid)))
        .execute(&connection)
        .expect("Error deleting notification");

    trace!("Deleted {} notifications", num_deleted);
}



pub fn delete_notification(policy_uuid: String, notification_id: i32) {
	// Create new diesel connection
	let connection = establish_connection();

	let num_deleted = diesel::delete(notifications::table.filter(notifications::policy_uuid.eq(policy_uuid))
												 		 .filter(notifications::notification_id.eq(notification_id)))
        .execute(&connection)
        .expect("Error deleting notification");

    trace!("Deleted {} notifications", num_deleted);
}


/****************************************************************************************
 * RuleNotification
 ****************************************************************************************/

pub fn add_rule_notification(policy_uuid: String, rule_num: i32, notification_id: i32) -> DatabaseRuleNotification {
	// Create new diesel connection
	let connection = establish_connection();
	let primary_key_policy_uuid = &policy_uuid;

	let new_rule_notification = DatabaseRuleNotification {
        policy_uuid: primary_key_policy_uuid.to_string(),
    	rule_num: rule_num,
		notification_id: notification_id,
    };

	// add a new policy if nothing is already there
	let results = rulenotifications::table.filter(rulenotifications::policy_uuid.eq(policy_uuid))
										.filter(rulenotifications::rule_num.eq(rule_num))
										.filter(rulenotifications::notification_id.eq(notification_id))
        .limit(QUERY_LIMIT)
        .load::<DatabaseRuleNotification>(&connection)
        .expect("Error loading rulenotifications");

	if results.len() > 0 {
		trace!("Rule notification already added");
		new_rule_notification	
	} else {
		trace!("Adding 1 rulenotification");
		diesel::insert_into(rulenotifications::table)
		    .values(&new_rule_notification)
		    .get_result(&connection)
		    .expect("Error saving new database rulenotification")
	}
}



pub fn get_rule_notifications_from_policy_uuid(policy_uuid: String) -> Vec<DatabaseRuleNotification> {
	// Create new diesel connection
	let connection = establish_connection();

	let results = rulenotifications::table.filter(rulenotifications::policy_uuid.eq(policy_uuid))
        .limit(QUERY_LIMIT)
        .load::<DatabaseRuleNotification>(&connection)
        .expect("Error loading rulenotifications");

    trace!("Getting {} rulenotifcations", results.len());
	results
}



pub fn get_rule_notifications(policy_uuid: String, rule_num: i32) -> Vec<DatabaseRuleNotification> {
	// Create new diesel connection
	let connection = establish_connection();

	let results = rulenotifications::table.filter(rulenotifications::policy_uuid.eq(policy_uuid))
												.filter(rulenotifications::rule_num.eq(rule_num))
        .limit(QUERY_LIMIT)
        .load::<DatabaseRuleNotification>(&connection)
        .expect("Error loading rulenotifications");

    trace!("Getting {} rulenotifcations", results.len());
	results
}



pub fn delete_rule_notifications(policy_uuid: String) {
	// Create new diesel connection
	let connection = establish_connection();

	let num_deleted = diesel::delete(rulenotifications::table.filter(rulenotifications::policy_uuid.eq(policy_uuid)))
        .execute(&connection)
        .expect("Error deleting rulenotifications");

    trace!("Deleted {} rulenotifications", num_deleted);
}



pub fn delete_rule_notification(policy_uuid: String, rule_num: i32, notification_id: i32) {
	// Create new diesel connection
	let connection = establish_connection();

	let num_deleted = diesel::delete(rulenotifications::table.filter(rulenotifications::policy_uuid.eq(policy_uuid))
												.filter(rulenotifications::rule_num.eq(rule_num))
												.filter(rulenotifications::notification_id.eq(notification_id)))
        .execute(&connection)
        .expect("Error deleting rulenotifications");

    trace!("Deleted {} rulenotifications", num_deleted);
}


/****************************************************************************************
 * Tags
 ****************************************************************************************/


pub fn add_tag(policy_uuid: String, tag_name: String) -> DatabaseTag {
	// Create new diesel connection
	let connection = establish_connection();
	let primary_key_policy_uuid = &policy_uuid;
	let primary_key_tag_name = &tag_name;

	let new_tag = DatabaseTag {
        policy_uuid: primary_key_policy_uuid.to_string(),
    	tag_name: primary_key_tag_name.to_string(),
    };

	// add a new policy if nothing is already there
	let results = tags::table.filter(tags::policy_uuid.eq(policy_uuid))
									  .filter(tags::tag_name.eq(tag_name))
        .limit(QUERY_LIMIT)
        .load::<DatabaseTag>(&connection)
        .expect("Error loading tags");

	if results.len() > 0 {
		trace!("Tag already added");
		new_tag
	} else {
		trace!("Adding 1 tag");
		diesel::insert_into(tags::table)
		    .values(&new_tag)
		    .get_result(&connection)
		    .expect("Error saving new database tag")
	}
}



pub fn get_tags_from_policy_uuid(policy_uuid: String) -> Vec<DatabaseTag> {
	// Create new diesel connection
	let connection = establish_connection();

	let results = tags::table.filter(tags::policy_uuid.eq(policy_uuid))
        .limit(QUERY_LIMIT)
        .load::<DatabaseTag>(&connection)
        .expect("Error loading tags");

    trace!("Getting {} tags", results.len());
	results
}



pub fn delete_tag(policy_uuid: String, tag_name: String) {
	// Create new diesel connection
	let connection = establish_connection();

	let num_deleted = diesel::delete(tags::table.filter(tags::policy_uuid.eq(policy_uuid))
														 .filter(tags::tag_name.eq(tag_name)))
        .execute(&connection)
        .expect("Error deleting tags");

    trace!("Deleted {} tags", num_deleted);
}



pub fn delete_tags_from_policy_uuid(policy_uuid: String) {
	// Create new diesel connection
	let connection = establish_connection();

	let num_deleted = diesel::delete(tags::table.filter(tags::policy_uuid.eq(policy_uuid)))
        .execute(&connection)
        .expect("Error deleting tags");

    trace!("Deleted {} tags", num_deleted);
}




