use core::hash::Hash;
use core::hash::Hasher;
use std::collections::HashMap;
use std::fmt;
use std::net::Ipv4Addr;
use std::net::Ipv6Addr;

use dyn_clone::DynClone;
use uuid::Uuid;

use super::schema::*;
use super::datastore;

/****************************************************************************************
 * Database
 ****************************************************************************************/

#[derive(Queryable, Insertable, AsChangeset, Clone, Debug)]
#[table_name="policies"]
pub struct DatabasePolicy {
    pub policy_uuid: String,
    pub rev_num: i32,
    pub certificate: Option<String>,
}

#[derive(Queryable, Insertable, AsChangeset, Clone, Debug)]
#[table_name="clients"]
pub struct DatabaseClient {
    pub client_id: i32,
    pub ipv4_addr: Option<String>,
	pub ipv6_addr: Option<String>,
    pub hostname: String,
}

#[derive(Insertable, AsChangeset, Clone, Debug)]
#[table_name="clients"]
pub struct NewDatabaseClient {
    pub ipv4_addr: Option<String>,
	pub ipv6_addr: Option<String>,
    pub hostname: String,
}

#[derive(Queryable, Insertable, AsChangeset, Clone, Debug)]
#[table_name="policyclients"]
pub struct DatabasePolicyClient {
    pub policy_uuid: String,
    pub client_id: i32,
}

#[derive(Queryable, Insertable, AsChangeset, Clone, Debug, Identifiable)]
#[table_name="rules"]
#[primary_key(policy_uuid, rule_num)]
pub struct DatabaseRule {
    pub policy_uuid: String,
    pub rule_num: i32,
    pub action: String,
    pub ipv4_remote_addr: Option<String>,
    pub ipv6_remote_addr: Option<String>,
    pub tcp_remote_port: Option<i32>,
    pub udp_remote_port: Option<i32>,
}

#[derive(Queryable, Insertable, AsChangeset, Clone, Debug, Identifiable)]
#[table_name="notifications"]
#[primary_key(policy_uuid, notification_id)]
pub struct DatabaseNotification {
	pub policy_uuid: String,
    pub notification_id: i32,
    pub email: Option<String>,
    pub rest_api: Option<String>,
}

#[derive(Queryable, Insertable, AsChangeset, Clone, Debug)]
#[table_name="rulenotifications"]
pub struct DatabaseRuleNotification {
	pub policy_uuid: String,
	pub rule_num: i32,
    pub notification_id: i32,
}

#[derive(Queryable, Insertable, AsChangeset, Clone, Debug)]
#[table_name="tags"]
pub struct DatabaseTag {
	pub policy_uuid: String,
    pub tag_name: String,
}





/****************************************************************************************
 * Firewall
 ****************************************************************************************/


/*
 * A policy contains multiple rules and tags that apply to a single client and a certificate to use for validating any future updates to the policy.
 */
pub struct Policy {
    pub uuid: Uuid,
    pub client: FirewallClient,
    pub revision: u32,
    pub certificate: Option<String>,
    pub rules: Vec<Rule>,
    pub tags: Vec<String>,
}

impl Policy {
    pub fn new(uuid: Uuid, revision: u32, client: FirewallClient) -> Policy {
        Policy {
            uuid,
            client,
            revision,
            certificate: None,
            rules: Vec::new(),
            tags: Vec::new(),
        }
    }

    pub fn add_tag(&mut self, tag: String) {
        if !self.tags.contains(&tag) {
            self.tags.push(tag);
        }
    }

    pub fn add_rule(&mut self, rule: Rule) {
        self.rules.push(rule);
    }

    pub fn set_certificate(&mut self, cert: String) {
        self.certificate = Some(cert);
    }

    pub fn process(&self, flow: *const Flow, bookmark: &usize) -> (MatchStatus, usize) {
        let mut i: usize = 0;
        for rule in &self.rules[*bookmark..] {
            let status: MatchStatus = rule.process(flow);
            match status {
                MatchStatus::MatchPass | MatchStatus::MatchDrop | MatchStatus::MatchReject => {
                    return (status, 0);
                }
                MatchStatus::NoMatch => {
                    i += 1;
                    continue;
                }
                MatchStatus::NotYetKnown => {
                    return (MatchStatus::NotYetKnown, i)
                }
            }
        }
        return (MatchStatus::NoMatch, 0);
    }
}

impl PartialEq for Policy {
    fn eq(&self, other: &Policy) -> bool {
        self.uuid == other.uuid
    }
}

impl Eq for Policy {}

impl Clone for Policy{
    fn clone(&self) -> Policy {
        let mut new_policy = Policy::new(self.uuid, self.revision, self.client.clone());
        new_policy.certificate = self.certificate.clone();
        for rule in &self.rules {
            new_policy.add_rule((*rule).clone());
        }
        for tag in &self.tags {
            new_policy.add_tag((*tag).clone());
        }
        return new_policy;
    }
}

impl Hash for Policy {
    fn hash<H>(&self, state: &mut H)
    where
        H: Hasher,
    {
        self.uuid.hash(state);
    }
}

/*
 * A firewall client.
 */
pub struct FirewallClient {
    pub ipv4_addr: Option<Ipv4Addr>,
    pub ipv6_addr: Option<Ipv6Addr>,
    pub hostname: String
}

impl FirewallClient {
    pub fn new(ipv4_addr: Option<Ipv4Addr>, ipv6_addr: Option<Ipv6Addr>, hostname: String) -> FirewallClient {
        FirewallClient {
            ipv4_addr,
            ipv6_addr,
            hostname
        }
    }
}

impl PartialEq for FirewallClient {
    fn eq(&self, other: &FirewallClient) -> bool {
        self.ipv4_addr == other.ipv4_addr && self.ipv6_addr == other.ipv6_addr && self.hostname == other.hostname
    }
}

impl Eq for FirewallClient {}

impl Clone for FirewallClient{
    fn clone(&self) -> FirewallClient {
        FirewallClient::new(self.ipv4_addr.clone(), self.ipv6_addr.clone(), self.hostname.clone())
    }
}

/*
 * A rule within a policy.
 * A rule contains an action and multiple selectors, all of which must be matched by a flow for the flow to match the rule.
 * A rule also contains notifications that should be sent when the rule is matched.
 */
pub struct Rule {
	pub action: String,
    pub selectors: Vec<Box<dyn Selector>>,
    pub notifications: Vec<Box<dyn Notification>>,
}

impl Rule {
    pub fn new(action: String) -> Rule {
        Rule {
            action,
            selectors: Vec::new(),
            notifications: Vec::new(),
        }
    }

    pub fn add_selector(&mut self, selector_name: String, selector_value: String) {
        let mut selector: Option<Box<dyn Selector>> = None;
        if selector_name == "ipv4-remote-addr" {
            match selector_value.parse::<Ipv4Addr>() {
                Ok(ip) => {
                    selector = Some(Box::new(Ipv4Address {
                        address: ip
                    }));
                }
                Err(_err) => {
                    warn!("Bad value {} for selector {}", selector_value, selector_name);
                    return;
                }
            }
        } else if selector_name == "ipv6-remote-addr" {
            match selector_value.parse::<Ipv6Addr>() {
                Ok(ip) => {
                    selector = Some(Box::new(Ipv6Address {
                        address: ip
                    }));
                }
                Err(_err) => {
                    warn!("Bad value {} for selector {}", selector_value, selector_name);
                    return;
                }
            }
        } else if selector_name == "tcp-remote-port" {
            match selector_value.parse::<u16>() {
                Ok(port) => {
                    selector = Some(Box::new(TcpPort {
                        port
                    }));
                }
                Err(_err) => {
                    warn!("Bad value {} for selector {}", selector_value, selector_name);
                    return;
                }
            }
        } else if selector_name == "udp-remote-port" {
            match selector_value.parse::<u16>() {
                Ok(port) => {
                    selector = Some(Box::new(UdpPort {
                        port
                    }));
                }
                Err(_err) => {
                    warn!("Bad value {} for selector {}", selector_value, selector_name);
                    return;
                }
            }
        }
        if selector.is_some() {
            self.selectors.push(selector.unwrap());
        } else {
            warn!("Unknown selector {}", selector_name);
        }
    }

	pub fn add_notification(&mut self, notification_name: String, notification_value: String) {
        let mut notification: Option<Box<dyn Notification>> = None;
        if notification_name == "email" {
            match notification_value.parse::<String>() {
                Ok(email) => {
                    notification = Some(Box::new(Email {
                        email: email
                    }));
                }
                Err(_err) => {
                    warn!("Bad value {} for notification {}", notification_value, notification_name);
                    return;
                }
            }
        } else if notification_name == "rest-api" {
            match notification_value.parse::<String>() {
                Ok(rest_api) => {
                    notification = Some(Box::new(RestApiCall {
                        rest_api: rest_api
                    }));
                }
                Err(_err) => {
                    warn!("Bad value {} for notification {}", notification_value, notification_name);
                    return;
                }
            }
        }
		if notification.is_some() {
            self.notifications.push(notification.unwrap());
        } else {
            warn!("Unknown selector {}", notification_name);
        }
	}

    /*
     * Check if the specified flow matches this rule and return a MatchStatus.
     * For a flow to match this rule, it must match all selectors in this rule.
     */
    pub fn process(&self, flow: *const Flow) -> MatchStatus {
        for selector in &self.selectors {
            let status: SelectorMatchStatus = selector.process(flow);
            match status {
                SelectorMatchStatus::NoMatch => {
                    return MatchStatus::NoMatch;
                }
                SelectorMatchStatus::NotYetKnown => {
                    return MatchStatus::NotYetKnown;
                }
                SelectorMatchStatus::Match => {
                    continue;
                }
            }
        }
        for notification in &self.notifications {
            if notification.my_type() == "email" {
                datastore::notify_email(notification.get_email());
                } else if notification.my_type() == "rest-api" {
                datastore::notify_api(notification.get_rest_api());
                }
        }
       if self.action == "pass" {
            return MatchStatus::MatchPass;
        } else if self.action == "drop" {
            return MatchStatus::MatchDrop;
        } else if self.action == "reject" {
            return MatchStatus::MatchReject;
        } else {
            panic!();
        }
    }
}

unsafe impl Send for Rule {}

impl Clone for Rule {
    fn clone(&self) -> Rule {
        let mut new_rule = Rule::new(self.action.clone());
        for selector in &self.selectors {
            new_rule.selectors.push(dyn_clone::clone_box(&**selector));
        }
        for notification in &self.notifications {
            new_rule.notifications.push(dyn_clone::clone_box(&**notification));
        }
        return new_rule;
    }
}

pub enum SelectorMatchStatus {
    Match,
    NoMatch,
    NotYetKnown,
}

pub enum MatchStatus {
    MatchPass,
    MatchDrop,
    MatchReject,
    NoMatch,
    NotYetKnown,
}

impl fmt::Display for MatchStatus {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        match *self {
            MatchStatus::MatchPass => write!(f, "ACCEPT"),
            MatchStatus::MatchDrop => write!(f, "DROP"),
            MatchStatus::MatchReject => write!(f, "REJECT"),
            MatchStatus::NoMatch => write!(f, "no match"),
            MatchStatus::NotYetKnown => write!(f, "not yet known"),
        }
    }
}

#[derive(Debug)]
pub enum FlowDirection {
    Outbound,
    Inbound,
}

impl fmt::Display for FlowDirection {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{:?}", self)
    }
}

/*
 * A network flow.
 */
pub struct Flow {
    pub client: Option<FirewallClient>,
    pub direction: Option<FlowDirection>,
    pub match_status: MatchStatus,
    pub match_state: HashMap<Policy, usize>,
    pub stream: Vec<u8>,
    pub ipv4_source: Option<Ipv4Addr>,
    pub ipv4_dest: Option<Ipv4Addr>,
    pub ipv6_source: Option<Ipv6Addr>,
    pub ipv6_dest: Option<Ipv6Addr>,
    pub ip_proto: u8,
    pub source_port: u16,
    pub dest_port: u16,
    pub http_host: Option<String>,
}

impl Flow {
    pub fn new(ipv4_source: Option<Ipv4Addr>, ipv4_dest: Option<Ipv4Addr>,
            ipv6_source: Option<Ipv6Addr>, ipv6_dest: Option<Ipv6Addr>,
            ip_proto: u8, source_port: u16, dest_port: u16) -> Flow {
        Flow {
            client: None,
            direction: None,
            match_status: MatchStatus::NotYetKnown,
            match_state: HashMap::new(),
            stream: Vec::new(),
            ipv4_source,
            ipv4_dest,
            ipv6_source,
            ipv6_dest,
            ip_proto,
            source_port,
            dest_port,
            http_host: None,
        }
    }
}

/****************************************************************************************
 * Notifications
 ****************************************************************************************/

pub trait Notification: DynClone {
	/*
     * Return a string of this Selector's type.
     */
	fn my_type(&self) -> &'static str {
        // This method should always be overridden
        unimplemented!();
    }
	// These functions don't need to implemented for every object
	fn get_email(&self) -> String {
        unimplemented!();
    }
	fn get_rest_api(&self) -> String {
        unimplemented!();
    }
}

#[derive(Clone)]
pub struct Email {
	email: String
}

impl Notification for Email {
	fn my_type(&self) -> &'static str {
        return "email";
    }
	
	fn get_email(&self) -> String {
        return self.email.to_string();
    }
}

#[derive(Clone)]
pub struct RestApiCall {
	rest_api: String
}

impl Notification for RestApiCall {
	fn my_type(&self) -> &'static str {
        return "rest-api";
    }

	fn get_rest_api(&self) -> String {
        return self.rest_api.to_string();
    }
}

/****************************************************************************************
 * Selectors
 ****************************************************************************************/

 /*
  * A selector within a rule.
  */
pub trait Selector: DynClone {
	/*
     * Return a string of this Selector's type.
     */
	fn my_type(&self) -> &'static str {
        // This method should always be overridden
        unimplemented!();
    }
    /*
     * Check if the specified flow matches this selector.
     */
    fn process(&self, flow: *const Flow) -> SelectorMatchStatus {
        // This method should always be overridden
        unimplemented!();
    }

	// These functions don't need to implemented for every object
	fn get_ipv4_address(&self) -> Ipv4Addr {
        unimplemented!();
    }
	fn get_ipv6_address(&self) -> Ipv6Addr {
        unimplemented!();
    }
	fn get_port(&self) -> u16 {
        unimplemented!();
    }
}

#[derive(Clone, Copy)]
pub struct Ipv4Address {
    address: Ipv4Addr
}

impl Selector for Ipv4Address {
	fn my_type(&self) -> &'static str {
        return "ipv4-remote-address";
    }
	fn get_ipv4_address(&self) -> Ipv4Addr {
        return self.address;
    }
    fn process(&self, flow: *const Flow) -> SelectorMatchStatus {
        unsafe {
            match (*flow).direction.as_ref().unwrap() {
                FlowDirection::Outbound => {
                    if (*flow).ipv4_dest.is_some() && (*flow).ipv4_dest.unwrap() == self.address {
                        return SelectorMatchStatus::Match;
                    }
                }
                FlowDirection::Inbound => {
                    if (*flow).ipv4_source.is_some() && (*flow).ipv4_source.unwrap() == self.address {
                        return SelectorMatchStatus::Match;
                    }
                }
            }
            return SelectorMatchStatus::NoMatch;
        }
    }
}

#[derive(Clone, Copy)]
pub struct Ipv6Address {
    address: Ipv6Addr
}

impl Selector for Ipv6Address {
	fn my_type(&self) -> &'static str {
        return "ipv6-remote-address";
    }
	fn get_ipv6_address(&self) -> Ipv6Addr {
        return self.address;
    }
    fn process(&self, flow: *const Flow) -> SelectorMatchStatus {
        unsafe {
            match (*flow).direction.as_ref().unwrap() {
                FlowDirection::Outbound => {
                    if (*flow).ipv6_dest.is_some() && (*flow).ipv6_dest.unwrap() == self.address {
                        return SelectorMatchStatus::Match;
                    }
                }
                FlowDirection::Inbound => {
                    if (*flow).ipv6_source.is_some() && (*flow).ipv6_source.unwrap() == self.address {
                        return SelectorMatchStatus::Match;
                    }
                }
            }
            return SelectorMatchStatus::NoMatch;
        }
    }
}

#[derive(Clone, Copy)]
pub struct TcpPort {
    port: u16
}

impl Selector for TcpPort {
	fn my_type(&self) -> &'static str {
        return "tcp-remote-port";
    }
	fn get_port(&self) -> u16 {
        return self.port;
    }
    fn process(&self, flow: *const Flow) -> SelectorMatchStatus {
        unsafe {
            if (*flow).ip_proto != 6 {
                return SelectorMatchStatus::NoMatch;
            }
            match (*flow).direction.as_ref().unwrap() {
                FlowDirection::Outbound => {
                    if (*flow).dest_port == self.port {
                        return SelectorMatchStatus::Match;
                    } else {
                        return SelectorMatchStatus::NoMatch;
                    }
                }
                FlowDirection::Inbound => {
                    if (*flow).source_port == self.port {
                        return SelectorMatchStatus::Match;
                    } else {
                        return SelectorMatchStatus::NoMatch;
                    }
                }
            }
        }
    }
}

#[derive(Clone, Copy)]
pub struct UdpPort {
    port: u16
}

impl Selector for UdpPort {
	fn my_type(&self) -> &'static str {
        return "udp-remote-port";
    }
	fn get_port(&self) -> u16 {
        return self.port;
    }
    fn process(&self, flow: *const Flow) -> SelectorMatchStatus {
        unsafe {
            if (*flow).ip_proto != 17 {
                return SelectorMatchStatus::NoMatch;
            }
            match (*flow).direction.as_ref().unwrap() {
                FlowDirection::Outbound => {
                    if (*flow).dest_port == self.port {
                        return SelectorMatchStatus::Match;
                    } else {
                        return SelectorMatchStatus::NoMatch;
                    }
                }
                FlowDirection::Inbound => {
                    if (*flow).source_port == self.port {
                        return SelectorMatchStatus::Match;
                    } else {
                        return SelectorMatchStatus::NoMatch;
                    }
                }
            }
        }
    }
}

#[derive(Clone, Copy)]
pub struct DnsQuestion {

}

impl Selector for DnsQuestion {}

#[derive(Clone)]
pub struct HttpHostHeader {
    hostname: String
}

impl Selector for HttpHostHeader {
    fn process(&self, flow: *const Flow) -> SelectorMatchStatus {
        unsafe {
            if (*flow).ip_proto != 6 || (*flow).dest_port != 443 {
                return SelectorMatchStatus::NoMatch;
            }
            if (*flow).http_host.is_some() && *(*flow).http_host.as_ref().unwrap() == self.hostname {
                return SelectorMatchStatus::Match;
            } else {
                return SelectorMatchStatus::NotYetKnown;
            }
        }
    }
}
