use std::sync::mpsc;

const IPTABLES_CHAIN: &str = "FORWARD";
const NFQUEUE_NUM: u16 = 26;

static mut queue: Option<nfqueue::Queue<State>> = None;

struct State {
    count: u32,
    tx: mpsc::Sender<Vec<u8>>,
    rx: mpsc::Receiver<nfqueue::Verdict>
}

impl State {
    pub fn new(tx: mpsc::Sender<Vec<u8>>, rx: mpsc::Receiver<nfqueue::Verdict>) -> State {
        State {count: 0, tx, rx}
    }
}

pub fn run(tx: mpsc::Sender<Vec<u8>>, rx: mpsc::Receiver<nfqueue::Verdict>) {
    let mut q = nfqueue::Queue::new(State::new(tx, rx));
    q.open();
    q.unbind(libc::AF_INET); // ignore result, failure is not critical here
    let rc = q.bind(libc::AF_INET);
    assert!(rc == 0);
    q.create_queue(NFQUEUE_NUM, nfqueue_callback);
    q.set_mode(nfqueue::CopyMode::CopyPacket, 0xffff);
    let mut iptv4: iptables::IPTables = iptables::new(false).unwrap();
    let mut iptv6: iptables::IPTables = iptables::new(true).unwrap();
    iptables_setup(&mut iptv4, &mut iptv6);
    unsafe {
        queue = Some(q);
        info!("Netfilterqueue thread entering main loop.");
        queue.as_ref().unwrap().run_loop();
        info!("Netfilterqueue thread exited main loop.");
        queue.as_mut().unwrap().destroy_queue();
        queue.as_mut().unwrap().close();
    }
    iptables_teardown(iptv4, iptv6);
}

fn iptables_setup(iptv4: &mut iptables::IPTables, iptv6: &mut iptables::IPTables) {
    iptv4.append("filter", IPTABLES_CHAIN, format!("-p tcp --dport 80 -j NFQUEUE --queue-num {}", NFQUEUE_NUM).as_str()).expect("Failed to add iptables rule");
    // iptv4.append("filter", IPTABLES_CHAIN, format!("-p tcp --dport 443 -j NFQUEUE --queue-num {}", NFQUEUE_NUM).as_str()).expect("Failed to add iptables rule");
    iptv4.append("filter", IPTABLES_CHAIN, format!("-p udp --dport 53 -j NFQUEUE --queue-num {}", NFQUEUE_NUM).as_str()).expect("Failed to add iptables rule");
    iptv4.append("filter", IPTABLES_CHAIN, format!("-p tcp --dport 23 -j NFQUEUE --queue-num {}", NFQUEUE_NUM).as_str()).expect("Failed to add iptables rule");
    iptv4.append("filter", IPTABLES_CHAIN, format!("-p udp --dport 17 -j NFQUEUE --queue-num {}", NFQUEUE_NUM).as_str()).expect("Failed to add iptables rule");
    iptv6.append("filter", IPTABLES_CHAIN, format!("-p tcp --dport 80 -j NFQUEUE --queue-num {}", NFQUEUE_NUM).as_str()).expect("Failed to add iptables rule");
    iptv6.append("filter", IPTABLES_CHAIN, format!("-p tcp --dport 443 -j NFQUEUE --queue-num {}", NFQUEUE_NUM).as_str()).expect("Failed to add iptables rule");
    iptv6.append("filter", IPTABLES_CHAIN, format!("-p udp --dport 53 -j NFQUEUE --queue-num {}", NFQUEUE_NUM).as_str()).expect("Failed to add iptables rule");
    iptv6.append("filter", IPTABLES_CHAIN, format!("-p tcp --dport 23 -j NFQUEUE --queue-num {}", NFQUEUE_NUM).as_str()).expect("Failed to add iptables rule");
    iptv6.append("filter", IPTABLES_CHAIN, format!("-p udp --dport 17 -j NFQUEUE --queue-num {}", NFQUEUE_NUM).as_str()).expect("Failed to add iptables rule");
    iptv4.append("nat", "POSTROUTING", "--out-interface eth1 -j MASQUERADE").expect("Failed to add iptables rule");
    iptv4.append("filter", "FORWARD", "--in-interface eth0 -j ACCEPT").expect("Failed to add iptables rule");
    iptv6.append("nat", "POSTROUTING", "--out-interface eth1 -j MASQUERADE").expect("Failed to add iptables rule");
    iptv6.append("filter", "FORWARD", "--in-interface eth0 -j ACCEPT").expect("Failed to add iptables rule");
}

fn iptables_teardown(iptv4: iptables::IPTables, iptv6: iptables::IPTables) {
    iptv6.delete("nat", "POSTROUTING", "--out-interface eth1 -j MASQUERADE").expect("Failed to remove iptables rule");
    iptv6.delete("filter", "FORWARD", "--in-interface eth0 -j ACCEPT").expect("Failed to remove iptables rule");
    iptv4.delete("nat", "POSTROUTING", "--out-interface eth1 -j MASQUERADE").expect("Failed to remove iptables rule");
    iptv4.delete("filter", "FORWARD", "--in-interface eth0 -j ACCEPT").expect("Failed to remove iptables rule");
    iptv4.delete("filter", IPTABLES_CHAIN, format!("-p tcp --dport 80 -j NFQUEUE --queue-num {}", NFQUEUE_NUM).as_str()).expect("Failed to remove iptables rule");
    // iptv4.delete("filter", IPTABLES_CHAIN, format!("-p tcp --dport 443 -j NFQUEUE --queue-num {}", NFQUEUE_NUM).as_str()).expect("Failed to remove iptables rule");
    iptv4.delete("filter", IPTABLES_CHAIN, format!("-p udp --dport 53 -j NFQUEUE --queue-num {}", NFQUEUE_NUM).as_str()).expect("Failed to remove iptables rule");
    iptv4.delete("filter", IPTABLES_CHAIN, format!("-p tcp --dport 23 -j NFQUEUE --queue-num {}", NFQUEUE_NUM).as_str()).expect("Failed to remove iptables rule");
    iptv4.delete("filter", IPTABLES_CHAIN, format!("-p udp --dport 17 -j NFQUEUE --queue-num {}", NFQUEUE_NUM).as_str()).expect("Failed to remove iptables rule");
    iptv6.delete("filter", IPTABLES_CHAIN, format!("-p tcp --dport 80 -j NFQUEUE --queue-num {}", NFQUEUE_NUM).as_str()).expect("Failed to remove iptables rule");
    iptv6.delete("filter", IPTABLES_CHAIN, format!("-p tcp --dport 443 -j NFQUEUE --queue-num {}", NFQUEUE_NUM).as_str()).expect("Failed to remove iptables rule");
    iptv6.delete("filter", IPTABLES_CHAIN, format!("-p udp --dport 53 -j NFQUEUE --queue-num {}", NFQUEUE_NUM).as_str()).expect("Failed to remove iptables rule");
    iptv6.delete("filter", IPTABLES_CHAIN, format!("-p tcp --dport 23 -j NFQUEUE --queue-num {}", NFQUEUE_NUM).as_str()).expect("Failed to remove iptables rule");
    iptv6.delete("filter", IPTABLES_CHAIN, format!("-p udp --dport 17 -j NFQUEUE --queue-num {}", NFQUEUE_NUM).as_str()).expect("Failed to remove iptables rule");
}

fn nfqueue_callback(msg: &nfqueue::Message, state:&mut State) {
    // println!("Packet received [id: 0x{:x}]\n", msg.get_id());
    // println!(" -> msg: {}", msg);
    // println!("XML\n{}", msg.as_xml_str(&[nfqueue::XMLFormatFlags::XmlAll]).unwrap());
    state.count += 1;
    debug!("count: {}", state.count);
    state.tx.send(msg.get_payload().to_vec());
    match state.rx.recv() {
        Ok(verdict) => {
            msg.set_verdict(verdict);
        }
        Err(_) => {
            unsafe {
                queue.as_mut().unwrap().stop = true;
            }
        }
    }
}
