use std::convert::TryInto;
use std::fs;
use std::net::IpAddr;
use std::process::Command;
use std::sync::mpsc;
use std::thread;
use std::time::Duration;
use thread::sleep;

use nix::unistd::Pid;
use nix::sys::signal::{self, Signal};

const IPTABLES_CHAIN: &str = "FORWARD";
const NFLOG_NUM_TSHARK: u16 = 26;
const SSLKEYLOGFILE: &str = "sslkeys";
const TSHARK_PCAP_FILE: &str = "pcap";

static mut SHUTDOWN_FLAG: bool = false;

pub fn shutdown() {
    unsafe {
        SHUTDOWN_FLAG = true;
    }
}

pub fn run(tx: mpsc::Sender<TsharkRecord>) {
    let mut iptv4: iptables::IPTables = iptables::new(false).unwrap();
    let mut iptv6: iptables::IPTables = iptables::new(true).unwrap();
    iptables_setup(&mut iptv4, &mut iptv6);
    info!("Tshark thread entering main loop.");
    tshark_loop(tx);
    info!("Tshark thread exited main loop.");
    iptables_teardown(iptv4, iptv6);
}

fn iptables_setup(iptv4: &mut iptables::IPTables, iptv6: &mut iptables::IPTables) {
    iptv4.append("filter", IPTABLES_CHAIN, format!("-p tcp --dport 443 -j NFLOG --nflog-group {}", NFLOG_NUM_TSHARK).as_str()).expect("Failed to add iptables rule");
    iptv4.append("filter", IPTABLES_CHAIN, format!("-p tcp --sport 443 -j NFLOG --nflog-group {}", NFLOG_NUM_TSHARK).as_str()).expect("Failed to add iptables rule");
}

fn iptables_teardown(iptv4: iptables::IPTables, iptv6: iptables::IPTables) {
    iptv4.delete("filter", IPTABLES_CHAIN, format!("-p tcp --dport 443 -j NFLOG --nflog-group {}", NFLOG_NUM_TSHARK).as_str()).expect("Failed to remove iptables rule");
    iptv4.delete("filter", IPTABLES_CHAIN, format!("-p tcp --sport 443 -j NFLOG --nflog-group {}", NFLOG_NUM_TSHARK).as_str()).expect("Failed to remove iptables rule");
}

pub struct TsharkRecord {
    pub hostname: String,
    pub src: IpAddr,
    pub src_port: u16,
    pub dst: IpAddr,
    pub dst_port: u16
}

fn tshark_loop(tx: mpsc::Sender<TsharkRecord>) {
    loop {
        unsafe {
            if SHUTDOWN_FLAG {
                return;
            }
        }
        fs::remove_file(TSHARK_PCAP_FILE);
        let tshark = Command::new("tshark")
            .args(&["-i", format!("nflog:{}", NFLOG_NUM_TSHARK).as_str(), "-w", TSHARK_PCAP_FILE])
            .spawn().expect("Failed to launch tshark");
        sleep(Duration::from_millis(5000));
        signal::kill(Pid::from_raw(tshark.id().try_into().unwrap()), Signal::SIGTERM).unwrap();
        let output: std::process::Output = Command::new("tshark")
            .args(&["-r", TSHARK_PCAP_FILE,
                "-o", format!("ssl.keylog_file: {}", SSLKEYLOGFILE).as_str(),
                "-o", "ssl.desegment_ssl_records: TRUE",
                "-o", "ssl.desegment_ssl_application_data: TRUE",
                "-Y", "http",
                "-T", "fields",
                "-e", "http.host",
                "-e", "tcp.port",
                "-e", "ip.src",
                "-e", "ip.dst"])
            .output().expect("Failed to launch tshark");
        let output: &str = std::str::from_utf8(output.stdout.as_slice()).unwrap();
        for line in output.split("\n") {
            let mut tokens = line.split_whitespace();
            match tokens.next() {
                Some(hostname) => {
                    let (src_port, dst_port) = match tokens.next() {
                        Some(ports_str) => {
                            let mut ports = ports_str.split(",");
                            (ports.next(), ports.next())
                        }
                        None => (None, None)
                    };
                    let src: Option<IpAddr> = match tokens.next() {
                        Some(source_ip_str) => {
                            match source_ip_str.parse() {
                                Ok(source_ip) => Some(source_ip),
                                Err(_e) => None
                            }
                        }
                        None => None
                    };
                    let dst: Option<IpAddr> = match tokens.next() {
                        Some(destination_ip_str) => {
                            match destination_ip_str.parse() {
                                Ok(destination_ip) => Some(destination_ip),
                                Err(_e) => None
                            }
                        }
                        None => None
                    };
                    if src.is_some() && src_port.is_some() && dst.is_some() && dst_port.is_some() {
                        if tx.send(TsharkRecord {
                            hostname: hostname.to_string(),
                            src: src.unwrap(),
                            src_port: src_port.unwrap().parse().unwrap(),
                            dst: dst.unwrap(),
                            dst_port: dst_port.unwrap().parse().unwrap()
                        }).is_err() {
                            return;
                        }
                    }
                }
                None => {}
            }
        }
    }
}
